// 134.2Khz Animal Rfid Reader Module Lange Afstand EM4305 Oormerk Reader Rfid Module ISO11784/85 FDX-B Ttl uart
/*
  version
  1.0 added leading 0's
  1.1 ignore start byte AA or early stop BB in message
  1.2 solved issue noted by ASalewski

  v2
  2.1 adapted hexInDec based on Sensorslot https://github.com/SensorsIot/Animal-RFID-Reader

*/
/*
  Taidacent 134.2khz Animal RFID Reader Module Lange Afstand EM4305 Oormerk Reader RFID Module ISO11784/85 FDX-B TTL uart
  De werkende frequentie van de kaartlezer module is 134.2KHZ (fout 1KHZ)
  Module werkspanning: 5V, stroom is minder dan 50MA
  De kaartlezer module leest ondersteuning tags die voldoen aan ISO 11784/85 FDX-B formaat of compatibel RFID kaarten
  Zoals duif enkels, oor tags en geïnjecteerd glazen buis labels, etc.
  Kaart lezen afstand: oormerk is 6CM, voet ring en andere 4CM
  De kaartlezer module is een ttl-niveau seriële poort uitgang, die direct kan worden aangesloten op de seriële poort van de doel microcontroller, en de baudrate is 9600
  Bijvoorbeeld, de hexadecimale data van het dier tag is: 03 84 12 DB FA E7 D5
  (Onder hen Land ID: 03 84, Nationale ID: 12 DB FA E7 D5)
  Seriële poort protocol: AA 0F 08 00 03 84 12 DB FA E7 D5 81 BB
  Waar AA is de start code 0F 08 00 (de drie data zijn vaste) 03 84 (land) 12 DB FA E7 D5 (data van animal tag) 81 (exclusieve of check code) BB (end code)
  Checksum berekening: 0F(XOR)08(XOR)00(XOR)03(XOR) 84(XOR)12(XOR)DB(XOR) FA(XOR)E7(XOR) d5 = 81
*/

/*
  I2C lcd 16x2 screen:
  GND-> ground
  VCC-> +5V
  SDA-> A4
  SCL-> A5
*/



#include <SoftwareSerial.h>
#include <LiquidCrystal_I2C.h>

const int BUFFER_SIZE = 13;       // RFID DATA FRAME FORMAT: 1byte head (AA=170), 10byte data (3byte fixed + 2byte country + 5byte tag), 1byte checksum, 1byte tail (BB=187)
const int DATA_SIZE = 10;         // 3 byte fixed (0F 08 00) + 7byte data (2byte country + 5byte tag) (03 84 + 12 DB FA E7 D5)
const int DATA_FIXED_SIZE = 3;    // 3byte fixed (0F 08 00)
const int DATA_COUNTRY_SIZE = 2;  // 2byte country (example 03 84)
const int DATA_TAG_SIZE = 5;      // 5byte tag (example 12 DB FA E7 D5)
const int CHECKSUM_SIZE = 1;      // 1byte checksum (example 81)

SoftwareSerial ssrfid = SoftwareSerial(6, 5); // Rx, Tx

uint8_t buffer[BUFFER_SIZE]; // used to store an incoming data frame

int buffer_index = 0;

//initialize the liquid crystal library
//the first parameter is the I2C address
//the second parameter is how many rows are on your screen
//the third parameter is how many columns are on your screen
LiquidCrystal_I2C lcd(0x27, 16, 2);

const int LedPin = 13;      // led is connected to pin D13
long previousMillis = 0;    // backlight timer
long interval = 120000;     // interval to turn off the backlight

void setup() {
  Serial.begin(9600);

  ssrfid.begin(9600);
  ssrfid.listen();

  pinMode(LedPin, OUTPUT);

  //initialize lcd screen
  lcd.init();
  // turn on the backlight
  lcd.backlight();
  previousMillis = millis();   // Start backlight timer
  lcd.clear();                 // clear display and set cursor to (0, 0)
  lcd.print("134.2khz Animal");
  lcd.setCursor(0, 1);
  lcd.print("  RFID Reader");

  Serial.println("INIT DONE");
}

void loop() {

  if (ssrfid.available() > 0) {     // als data aanwezig
    bool call_extract_tag = false;

    int ssvalue = ssrfid.read(); // lees data

    if (ssvalue == -1) {         // no data was read
      return;
    }

    if (ssvalue == 170 && buffer_index == 0) {           // EM4305RFID found a tag => tag incoming
      //      buffer_index = 0;
      digitalWrite(LedPin, HIGH);   // led aan
    } else if (ssvalue == 187 && buffer_index == BUFFER_SIZE - 1) {    // tag has been fully transmitted
      call_extract_tag = true;      // extract tag at the end of the function call
    }

    if (buffer_index >= BUFFER_SIZE) {  // checking for a buffer overflow (It's very unlikely that an buffer overflow comes up!)
      Serial.println("Error: Buffer overflow detected!");
      lcd.clear();
      lcd.backlight();
      previousMillis = millis();   // Start backlight timer
      lcd.print("Buffer overflow");
      buffer_index = 0;
      return;
    }

    buffer[buffer_index++] = ssvalue; // everything is alright => copy current value to buffer
    Serial.print(ssvalue);
    Serial.print(" ");

    if (call_extract_tag == true) {
      if (buffer_index == BUFFER_SIZE) {
        extract_tag();
        digitalWrite(LedPin, LOW);
        buffer_index = 0;
      } else { // something is wrong... start again looking for preamble (value: 170)
        buffer_index = 0;
        return;
      }
    }
  }

  if (previousMillis + interval < millis()) {   // check backlight timer
    lcd.noBacklight();                          // als interval verstreken dan backlight uit
  }

}

void extract_tag() {          // analiseer data

  uint8_t msg_checksum = buffer[11]; // 1 byte
  uint8_t msg_data[DATA_SIZE];
  uint8_t msg_data_country[DATA_COUNTRY_SIZE];
  uint8_t msg_data_tag[DATA_TAG_SIZE];
  uint8_t countrymessage[2 * DATA_COUNTRY_SIZE]; // used to store ASCII code of hex data of country part
  uint8_t tagmessage[2 * DATA_TAG_SIZE]; // used tot store ASCII code of hex data of tag part
  long checksum = 0;
  long country = 0;
  long long tag = 0;
  long s;                 // second half of tag to show and determine leading 0
  int expo;               //exponent counter

  // test chip numbers, comment out if not testing
  // uint8_t buffer[BUFFER_SIZE] = {170, 15, 8, 0, 2, 16, 52, 39, 128, 22, 6, 150, 187};  // test known chip SKIP 528 224001005062
  // uint8_t buffer[BUFFER_SIZE] = {170, 15, 8, 0, 2, 16, 0, 29, 210, 159, 249, 0, 187};  // test chip 0, 29, 210, 159, 249 = 00 1D D2 9F F9 or DEC 500342777  (checksum not right)
  // uint8_t buffer[BUFFER_SIZE] = {170, 15, 8, 0, 2, 16, 0, 29, 252, 190, 25, 0, 187};  // 0, 29, 252, 190, 25 =  00 1D FC BE 19 or dec 503103001
  // end testing

  Serial.println("");
  Serial.print("Buffer (DEC): ");
  for (int i = 0; i < BUFFER_SIZE; i++) {
    Serial.print(buffer[i], DEC);
    Serial.print(" ");
  }

  Serial.println("");
  Serial.print("Buffer (HEX): ");
  for (int i = 0; i < BUFFER_SIZE; i++) {
    Serial.print(buffer[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  Serial.print("Buffer DATA (HEX): ");
  for (int i = 0; i < DATA_SIZE; i++) {
    msg_data[i] = buffer[i + 1];
    Serial.print(msg_data[i], HEX);
    checksum ^= msg_data[i];              // a = a XOR b
  }
  Serial.println("");

  Serial.print("Extracted Checksum (HEX): ");
  Serial.print(checksum, HEX);

  if (checksum == msg_checksum) { // compare calculated checksum to retrieved checksum
    Serial.print(" (OK)"); // calculated checksum corresponds to transmitted checksum!
  } else {
    Serial.print(" (NOT OK)"); // checksums do not match
  }
  Serial.println("");
  
  Serial.print("Buffer COUNTRY (HEX): ");
  for (int i = 0; i < DATA_COUNTRY_SIZE; i++) {
    msg_data_country[i] = buffer[i + 4];
    Serial.print(msg_data_country[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  // put buffer country in reverse into country message
  for (int i = 0; i < DATA_COUNTRY_SIZE ; i++) {
    String str = String(buffer[i + 4], HEX);
    if (str.length() == 1 && buffer[i + 6] - 48 < 10) {  // add leading 0 if < 10
      str = "0" + str;
    }
    str.toUpperCase();
    for (int k = 0; k < str.length(); k++) {
      byte x = str.charAt(k);
      countrymessage[2 * DATA_COUNTRY_SIZE - 1 - i * 2 - k] = x;
    }
  }

  country = hexInDec(countrymessage, 0, 2 * DATA_COUNTRY_SIZE);

  Serial.print("Country (DEC): ");
  Serial.println(country);

  Serial.print("Buffer TAG (HEX): ");
  for (int i = 0; i < DATA_TAG_SIZE; i++) {
    msg_data_tag[i] = buffer[i + 6];
    Serial.print(msg_data_tag[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  // Buffer TAG in reverse to tagmessage
  for (int i = 0; i < DATA_TAG_SIZE ; i++) {
    String str = String(buffer[i + 6], HEX);
    if (str.length() == 1 && buffer[i + 6] - 48 < 10) {  // add leading 0 if < 10
      str = "0" + str;
    }
    str.toUpperCase();
    for (int k = 0; k < str.length(); k++) {
      byte x = str.charAt(k);
      tagmessage[2 * DATA_TAG_SIZE - 1 - i * 2 - k] = x;
    }
  }

  tag = hexInDec(tagmessage, 0, 2 * DATA_TAG_SIZE);

  Serial.print("Tag (DEC): ");
  Serial.print(long(tag / 1000000));
  Serial.print(" ");
  //Serial.println(long(tag % 1000000));  // leading 0 are not printed
  s = long(tag % 1000000);
  for (long i = 100000; s < i && i > 1; i = i / 10) {   // print leading 0
    Serial.print("0");
  }
  Serial.println(s);

  Serial.println("--------");

  lcd.clear();
  lcd.backlight();
  previousMillis = millis();   // reset backlight timer
  lcd.print(country);
  if (checksum == msg_checksum) { // compare calculated checksum to retrieved checksum
    lcd.print(" (OK)"); // calculated checksum corresponds to transmitted checksum!
  } else {
    lcd.print(" (NOT OK)"); // checksums do not match
  }
  lcd.setCursor(0, 1);
  lcd.print(long(tag / 1000000));
  s = long(tag % 1000000);
  for (long i = 100000; s < i && i > 1; i = i / 10) { // print leading 0
    lcd.print("0");
  }
  lcd.print(long(tag % 1000000));

}

long long hexInDec(char message[], int beg , int len) {
  long long mult = 1;
  long long nbr = 0;
  byte nextInt;
  for (int i = beg; i < beg + len; i++) {
    nextInt = message[i];
    if (nextInt >= 48 && nextInt <= 57) nextInt = map(nextInt, 48, 57, 0, 9);
    if (nextInt >= 65 && nextInt <= 70) nextInt = map(nextInt, 65, 70, 10, 15);
    if (nextInt >= 97 && nextInt <= 102) nextInt = map(nextInt, 97, 102, 10, 15);
    nextInt = constrain(nextInt, 0, 15);
    nbr = nbr + (mult * nextInt);
    mult = mult * 16;
  }
  return nbr;
}
